# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import os
import re
from urllib.parse import unquote_plus
from urllib.parse import urlparse

from packageurl import PackageURL
from packageurl.contrib.route import NoRouteAvailable
from packageurl.contrib.route import Router

"""
This module helps build a PackageURL from an arbitrary URL.
This uses the a routing mechanism available in the route.py module.

In order to make it easy to use, it contains all the conversion functions
in this single Python script.
"""


purl_router = Router()


def url2purl(url):
    """
    Return a PackageURL inferred from the `url` string or None.
    """
    if url:
        try:
            return purl_router.process(url)
        except NoRouteAvailable:
            # If `url` does not fit in one of the existing routes,
            # we attempt to create a generic PackageURL for `url`
            return build_generic_purl(url)


get_purl = url2purl


def purl_from_pattern(type_, pattern, url):
    url = unquote_plus(url)
    compiled_pattern = re.compile(pattern, re.VERBOSE)
    match = compiled_pattern.match(url)

    if match:
        purl_data = {
            field: value for field, value in match.groupdict().items()
            if field in PackageURL._fields
        }
        return PackageURL(type_, **purl_data)


def register_pattern(type_, pattern, router=purl_router):
    """
    Register a pattern with its type.
    """
    def endpoint(url):
        return purl_from_pattern(type_, pattern, url)
    router.append(pattern, endpoint)


def get_path_segments(url):
    """
    Return a list of path segments from a `url` string.
    """
    path = unquote_plus(urlparse(url).path)
    segments = [seg for seg in path.split("/") if seg]
    return segments


def build_generic_purl(uri):
    """
    Return a PackageURL from `uri`, if `uri` is a parsable URL, or None

    `uri` is assumed to be a download URL, e.g. http://example.com/example.tar.gz
    """
    parsed_uri = urlparse(uri)
    if parsed_uri.scheme and parsed_uri.netloc and parsed_uri.path:
        # Get file name from `uri`
        uri_path_segments = get_path_segments(uri)
        file_name = uri_path_segments[-1]
        return PackageURL(
            type='generic',
            name=file_name,
            qualifiers={
                'download_url': uri
            }
        )


@purl_router.route('https?://registry.npmjs.*/.*',
                   'https?://registry.yarnpkg.com/.*',
                   'https?://(www\\.)?npmjs.*/package.*',
                   'https?://(www\\.)?yarnpkg.com/package.*')
def build_npm_purl(uri):
    # npm URLs are difficult to disambiguate with regex
    if '/package/' in uri:
        return build_npm_web_purl(uri)
    elif '/-/' in uri:
        return build_npm_download_purl(uri)
    else:
        return build_npm_api_purl(uri)


def build_npm_api_purl(uri):
    path = unquote_plus(urlparse(uri).path)
    segments = [seg for seg in path.split('/') if seg]

    if len(segments) != 2:
        return

    # /@invisionag/eslint-config-ivx
    if segments[0].startswith('@'):
        namespace = segments[0]
        name = segments[1]
        return PackageURL('npm', namespace, name)

    # /angular/1.6.6
    else:
        name = segments[0]
        version = segments[1]
        return PackageURL('npm', name=name, version=version)


def build_npm_download_purl(uri):
    path = unquote_plus(urlparse(uri).path)
    segments = [seg for seg in path.split('/') if seg and seg != '-']
    len_segments = len(segments)

    # /@invisionag/eslint-config-ivx/-/eslint-config-ivx-0.0.2.tgz
    if len_segments == 3:
        namespace, name, filename = segments

    # /automatta/-/automatta-0.0.1.tgz
    elif len_segments == 2:
        namespace = None
        name, filename = segments

    else:
        return

    base_filename, ext = os.path.splitext(filename)
    version = base_filename.split('-')[-1]

    return PackageURL('npm', namespace, name, version)


def build_npm_web_purl(uri):
    path = unquote_plus(urlparse(uri).path)
    if path.startswith('/package/'):
        path = path[9:]

    segments = [seg for seg in path.split('/') if seg]
    len_segments = len(segments)
    namespace = version = None

    # @angular/cli/v/10.1.2
    if len_segments == 4:
        namespace = segments[0]
        name = segments[1]
        version = segments[3]

    # express/v/4.17.1
    elif len_segments == 3:
        namespace = None
        name = segments[0]
        version = segments[2]

    # @angular/cli
    elif len_segments == 2:
        namespace = segments[0]
        name = segments[1]

    # express
    elif len_segments == 1 and len(segments) > 0 and segments[0][0] != '@':
        name = segments[0]

    else:
        return

    return PackageURL('npm', namespace, name, version)


@purl_router.route('https?://repo1.maven.org/maven2/.*',
                   'https?://central.maven.org/maven2/.*',
                   'maven-index://repo1.maven.org/.*')
def build_maven_purl(uri):
    path = unquote_plus(urlparse(uri).path)
    segments = [seg for seg in path.split('/') if seg and seg != 'maven2']

    if len(segments) < 3:
        return

    before_last_segment, last_segment = segments[-2:]
    has_filename = before_last_segment in last_segment

    filename = None
    if has_filename:
        filename = segments.pop()

    version = segments[-1]
    name = segments[-2]
    namespace = '.'.join(segments[:-2])
    qualifiers = {}

    if filename:
        name_version = '{}-{}'.format(name, version)
        _, _, classifier_ext = filename.rpartition(name_version)
        classifier, _, extension = classifier_ext.partition('.')
        if not extension:
            return

        qualifiers['classifier'] = classifier.strip('-')

        valid_types = ('aar', 'ear', 'mar', 'pom', 'rar', 'rpm',
                       'sar', 'tar.gz', 'war', 'zip')
        if extension in valid_types:
            qualifiers['type'] = extension

    return PackageURL('maven', namespace, name, version, qualifiers)


@purl_router.route('https?://rubygems.org/downloads/.*')
def build_rubygems_purl(uri):
    # We use a more general route pattern instead of using `rubygems_pattern`
    # below by itself because we want to capture all rubygems download URLs,
    # even the ones that are not completly formed. This helps prevent url2purl
    # from attempting to create a generic PackageURL from an invalid rubygems
    # download URL.

    # https://rubygems.org/downloads/jwt-0.1.8.gem
    rubygems_pattern = (
        r"^https?://rubygems.org/downloads/"
        r"(?P<name>.+)-(?P<version>.+)"
        r"(\.gem)$"
    )
    return purl_from_pattern('rubygems', rubygems_pattern, uri)


# https://pypi.python.org/packages/source/a/anyjson/anyjson-0.3.3.tar.gz
pypi_pattern = (
    r"(?P<name>.+)-(?P<version>.+)"
    r"\.(zip|tar.gz|tar.bz2|.tgz)$"
)

# This pattern can be found in the following locations:
# - wheel.wheelfile.WHEEL_INFO_RE
# - distlib.wheel.FILENAME_RE
# - setuptools.wheel.WHEEL_NAME
# - pip._internal.wheel.Wheel.wheel_file_re
wheel_file_re = re.compile(
    r"^(?P<namever>(?P<name>.+?)-(?P<version>.*?))"
    r"((-(?P<build>\d[^-]*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)"
    r"\.whl)$",
    re.VERBOSE
)


@purl_router.route('https?://.+python.+org/packages/.*')
def build_pypi_purl(uri):
    path = unquote_plus(urlparse(uri).path)
    last_segment = path.split('/')[-1]

    # /wheel-0.29.0-py2.py3-none-any.whl
    if last_segment.endswith('.whl'):
        match = wheel_file_re.match(last_segment)
        if match:
            return PackageURL(
                'pypi',
                name=match.group('name'),
                version=match.group('version'),
            )

    return purl_from_pattern('pypi', pypi_pattern, last_segment)


# http://nuget.org/packages/EntityFramework/4.2.0.0
# https://www.nuget.org/api/v2/package/Newtonsoft.Json/11.0.1
nuget_www_pattern = (
    r"^https?://.*nuget.org/(api/v2/)?packages?/"
    r"(?P<name>.+)/"
    r"(?P<version>.+)$"
)

register_pattern('nuget', nuget_www_pattern)


# https://api.nuget.org/v3-flatcontainer/newtonsoft.json/10.0.1/newtonsoft.json.10.0.1.nupkg
nuget_api_pattern = (
    r"^https?://api.nuget.org/v3-flatcontainer/"
    r"(?P<name>.+)/"
    r"(?P<version>.+)/"
    r".*(nupkg)$"  # ends with "nupkg"
)

register_pattern('nuget', nuget_api_pattern)


@purl_router.route('https?://.*sourceforge.net/project/.*')
def build_sourceforge_purl(uri):
    # We use a more general route pattern instead of using `sourceforge_pattern`
    # below by itself because we want to capture all sourceforge download URLs,
    # even the ones that do not fit `sourceforge_pattern`. This helps prevent
    # url2purl from attempting to create a generic PackageURL from a sourceforge
    # URL that we can't handle.

    # http://master.dl.sourceforge.net/project/libpng/zlib/1.2.3/zlib-1.2.3.tar.bz2
    sourceforge_pattern = (
        r"^https?://.*sourceforge.net/project/"
        r"(?P<namespace>([^/]+))/"  # do not allow more "/" segments
        r"(?P<name>.+)/"
        r"(?P<version>[0-9\.]+)/"  # version restricted to digits and dots
        r"(?P=name)-(?P=version).*"  # {name}-{version} repeated in the filename
        r"[^/]$"  # not ending with "/"
    )

    sourceforge_purl = purl_from_pattern('sourceforge', sourceforge_pattern, uri)

    if not sourceforge_purl:
        # We create a more generic PackageURL from `uri` if `uri` doesn't fit
        # `sourceforge_pattern`
        uri_path_segments = get_path_segments(uri)
        file_name = uri_path_segments[-1]
        sourceforge_purl = PackageURL(
            type='sourceforge',
            name=file_name,
            qualifiers={
                'download_url': uri
            }
        )

    return sourceforge_purl


# https://crates.io/api/v1/crates/rand/0.7.2/download
cargo_pattern = (
    r"^https?://crates.io/api/v1/crates/"
    r"(?P<name>.+)/(?P<version>.+)"
    r"(\/download)$"
)

register_pattern('cargo', cargo_pattern)


# https://raw.githubusercontent.com/volatilityfoundation/dwarf2json/master/LICENSE.txt
github_raw_content_pattern = (
    r"https?://raw.githubusercontent.com/(?P<namespace>[^/]+)/(?P<name>[^/]+)/"
    r"(?P<version>[^/]+)/(?P<subpath>.*)$"
)

register_pattern('github', github_raw_content_pattern)


@purl_router.route("https?://api.github\\.com/repos/.*")
def build_github_api_purl(url):
    """
    Return a PackageURL object from GitHub API `url`.
    For example:
    https://api.github.com/repos/nexB/scancode-toolkit/commits/40593af0df6c8378d2b180324b97cb439fa11d66
    https://api.github.com/repos/nexB/scancode-toolkit/
    and returns a `PackageURL` object
    """
    segments = get_path_segments(url)

    if not(len(segments) >= 3):
        return
    namespace = segments[1]
    name = segments[2]
    version = None

    # https://api.github.com/repos/nexB/scancode-toolkit/
    if len(segments) == 4 and segments[3] != 'commits':
        version = segments[3]

    # https://api.github.com/repos/nexB/scancode-toolkit/commits/40593af0df6c8378d2b180324b97cb439fa11d66
    if len(segments) == 5 and segments[3] == "commits":
        version = segments[4]

    return PackageURL(
        type='github', namespace=namespace, name=name, version=version
    )


# https://codeload.github.com/nexB/scancode-toolkit/tar.gz/v3.1.1
# https://codeload.github.com/berngp/grails-rest/zip/release/0.7
github_codeload_pattern = (
    r"https?://codeload.github.com/(?P<namespace>.+)/(?P<name>.+)/"
    r"(zip|tar.gz|tar.bz2|tgz)/(.*/)*"
    r"v?(?P<version>.+)$"
)

register_pattern('github', github_codeload_pattern)


@purl_router.route("https?://github\\.com/.*")
def build_github_purl(url):
    """
    Return a PackageURL object from GitHub `url`.
    """

    # https://github.com/nexB/scancode-toolkit/archive/v3.1.1.zip
    archive_pattern = (
        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
        r"/archive/(.*/)*"
        r"((?P=name)(-|_|@))?"
        r"v?(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
    )

    # https://github.com/downloads/mozilla/rhino/rhino1_7R4.zip
    download_pattern = (
        r"https?://github.com/downloads/(?P<namespace>.+)/(?P<name>.+)/"
        r"((?P=name)(-|@)?)?"
        r"v?(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
    )

    # https://github.com/pypa/get-virtualenv/raw/20.0.31/public/virtualenv.pyz
    raw_pattern = (
        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
        r"/raw/v?(?P<version>[^/]+)/(?P<subpath>.*)$"
    )

    # https://github.com/fanf2/unifdef/blob/master/unifdef.c
    blob_pattern = (
        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
        r"/blob/(?P<version>[^/]+)/(?P<subpath>.*)$"
    )

    releases_download_pattern = (
        r"https?://github.com/(?P<namespace>.+)/(?P<name>.+)"
        r"/releases/download/v?(?P<version>[^/]+)/.*$"
    )

    # https://github.com/pombredanne/schematics.git
    git_pattern = r"https?://github.com/(?P<namespace>.+)/(?P<name>.+).(git)"

    patterns = (
        archive_pattern,
        raw_pattern,
        blob_pattern,
        releases_download_pattern,
        download_pattern,
        git_pattern,
    )

    for pattern in patterns:
        matches = re.search(pattern, url)
        qualifiers = {}
        if matches:
            if pattern != releases_download_pattern:
                return purl_from_pattern(type_='github', pattern=pattern, url=url)
            qualifiers['download_url'] = url
            purl = purl_from_pattern(type_='github', pattern=pattern, url=url)
            return PackageURL(
                type=purl.type,
                name=purl.name,
                namespace=purl.namespace,
                version=purl.version,
                qualifiers=qualifiers
            )

    segments = get_path_segments(url)
    if not len(segments) >= 2:
        return

    namespace = segments[0]
    name = segments[1]
    version = None
    subpath = None

    # https://github.com/TG1999/fetchcode/master
    if len(segments) >= 3 and segments[2] != 'tree':
        version = segments[2]
        subpath = '/'.join(segments[3:])

    # https://github.com/TG1999/fetchcode/tree/master
    if len(segments) >= 4 and segments[2] == 'tree':
        version = segments[3]
        subpath = '/'.join(segments[4:])

    return PackageURL(
        type='github',
        namespace=namespace,
        name=name,
        version=version,
        subpath=subpath,
    )


@purl_router.route("https?://bitbucket\\.org/.*")
def build_bitbucket_purl(url):
    """
    Return a PackageURL object from BitBucket `url`.
    For example:
    https://bitbucket.org/TG1999/first_repo/src/master or
    https://bitbucket.org/TG1999/first_repo/src or
    https://bitbucket.org/TG1999/first_repo/src/master/new_folder
    """

    segments = get_path_segments(url)

    if not len(segments) >= 2:
        return
    namespace = segments[0]
    name = segments[1]

    bitbucket_download_pattern = (
        r"https?://bitbucket.org/"
        r"(?P<namespace>.+)/(?P<name>.+)/downloads/(?P<version>.+).(zip|tar.gz|tar.bz2|.tgz)"
    )
    matches = re.search(bitbucket_download_pattern, url)

    qualifiers = {}
    if matches:
        qualifiers['download_url'] = url
        return PackageURL(
            type='bitbucket',
            namespace=namespace,
            name=name,
            qualifiers=qualifiers
        )

    version = None
    subpath = None

    # https://bitbucket.org/TG1999/first_repo/new_folder/
    if len(segments) >= 3 and segments[2] != 'src':
        version = segments[2]
        subpath = '/'.join(segments[3:])

    # https://bitbucket.org/TG1999/first_repo/src/master/new_folder/
    if len(segments) >= 4 and segments[2] == 'src':
        version = segments[3]
        subpath = '/'.join(segments[4:])

    return PackageURL(
        type='bitbucket',
        namespace=namespace,
        name=name,
        version=version,
        subpath=subpath,
    )


@purl_router.route("https?://gitlab\\.com/.*")
def build_gitlab_purl(url):
    """
    Return a PackageURL object from Gitlab `url`.
    For example:
    https://gitlab.com/TG1999/firebase/-/tree/1a122122/views
    https://gitlab.com/TG1999/firebase/-/tree
    https://gitlab.com/TG1999/firebase/-/master
    https://gitlab.com/tg1999/Firebase/-/tree/master
    """
    segments = get_path_segments(url)

    if not len(segments) >= 2:
        return
    namespace = segments[0]
    name = segments[1]
    version = None
    subpath = None

    # https://gitlab.com/TG1999/firebase/master
    if (len(segments) >= 3) and segments[2] != '-' and segments[2] != 'tree':
        version = segments[2]
        subpath = '/'.join(segments[3:])

    # https://gitlab.com/TG1999/firebase/-/tree/master
    if len(segments) >= 5 and (segments[2] == '-' and segments[3] == 'tree'):
        version = segments[4]
        subpath = '/'.join(segments[5:])

    return PackageURL(
        type='gitlab',
        namespace=namespace,
        name=name,
        version=version,
        subpath=subpath,
    )


# https://hackage.haskell.org/package/a50-0.5/a50-0.5.tar.gz
hackage_pattern = (
    r"^https?://hackage.haskell.org/package/"
    r"(?P<name>.+)-(?P<version>.+)/"
    r"(?P=name)-(?P=version).*"
    r"[^/]$"
)

register_pattern('hackage', hackage_pattern)



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import json
import os
import re
import unittest

from packageurl import PackageURL
from packageurl import normalize
from packageurl import normalize_qualifiers


def create_test_function(description, purl, canonical_purl, is_invalid,
                         type, name, namespace, version, qualifiers, subpath,  # NOQA
                         test_func_prefix='test_purl_pkg_', **kwargs):
    """
    Return a new (test function, test_name) where the test_function closed on
    test arguments. If is_error is True the tests are expected to raise an
    Exception.
    """
    if is_invalid:

        def test_purl(self):
            try:
                PackageURL.from_string(purl)
                self.fail('Should raise a ValueError')
            except ValueError:
                pass

            try:
                PackageURL.from_string(canonical_purl)
                self.fail('Should raise a ValueError')
            except ValueError:
                pass

            try:
                PackageURL(type, namespace, name, version, qualifiers, subpath)
            except ValueError:
                pass

    else:

        def test_purl(self):
            # parsing the test canonical `purl` then re-building a `purl` from these
            # parsed components should return the test canonical `purl`
            cano = PackageURL.from_string(purl)
            assert canonical_purl == cano.to_string()

            # parsing the test `purl` should return the components parsed from the
            # test canonical `purl`
            parsed = PackageURL.from_string(canonical_purl)
            assert cano.to_dict() == parsed.to_dict()

            # parsing the test `purl` then re-building a `purl` from these parsed
            # components should return the test canonical `purl`
            assert canonical_purl == parsed.to_string()

            # building a `purl` from the test components should return the test
            # canonical `purl`
            built = PackageURL(type, namespace, name, version, qualifiers, subpath)
            assert canonical_purl == built.to_string()

    # create a good function name for use in test discovery
    if not description:
        description = purl
    if is_invalid:
        test_func_prefix += 'is_invalid_'
    test_name = python_safe_name(test_func_prefix + description)
    test_purl.__name__ = test_name
    test_purl.funcname = test_name
    return test_purl, test_name


def python_safe_name(s):
    """
    Return a name derived from string `s` safe to use as a Python function name.

    For example:
    >>> s = "not `\\a /`good` -safe name ??"
    >>> assert python_safe_name(s) == 'not_good_safe_name'
    """
    no_punctuation = re.compile(r'[\W_]', re.MULTILINE).sub
    s = s.lower()
    s = no_punctuation(' ', s)
    s = '_'.join(s.split())
    return s


class PurlTest(unittest.TestCase):
    pass


def build_tests(clazz=PurlTest, test_file='test-suite-data.json'):
    """
    Dynamically build test methods for each purl test found in the `test_file`
    JSON file and attach a test method to the `clazz` class.
    """
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')
    test_file = os.path.join(test_data_dir, test_file)

    with open(test_file) as tf:
        tests_data = json.load(tf)
    for items in tests_data:
        test_func, test_name = create_test_function(**items)
        # attach that method to the class
        setattr(clazz, test_name, test_func)


build_tests()


class NormalizePurlTest(unittest.TestCase):

    def test_normalize_qualifiers_as_string(self):
        qualifiers_as_dict = {
            'classifier': 'sources',
            'repository_url': 'repo.spring.io/release'
        }
        qualifiers_as_string = 'classifier=sources&repository_url=repo.spring.io/release'
        assert qualifiers_as_string == normalize_qualifiers(
            qualifiers_as_dict, encode=True)

    def test_normalize_qualifiers_as_dict(self):
        qualifiers_as_dict = {
            'classifier': 'sources',
            'repository_url': 'repo.spring.io/release'
        }
        qualifiers_as_string = 'classifier=sources&repository_url=repo.spring.io/release'
        assert qualifiers_as_dict == normalize_qualifiers(
            qualifiers_as_string, encode=False)

    def test_create_PackageURL_from_qualifiers_string(self):
        canonical_purl = 'pkg:maven/org.apache.xmlgraphics/batik-anim@1.9.1?classifier=sources&repository_url=repo.spring.io/release'
        type = 'maven'  # NOQA
        namespace = 'org.apache.xmlgraphics'
        name = 'batik-anim'
        version = '1.9.1'
        qualifiers_as_string = 'classifier=sources&repository_url=repo.spring.io/release'
        subpath = None

        purl = PackageURL(type, namespace, name, version, qualifiers_as_string,
                          subpath)
        assert canonical_purl == purl.to_string()

    def test_create_PackageURL_from_qualifiers_dict(self):
        canonical_purl = 'pkg:maven/org.apache.xmlgraphics/batik-anim@1.9.1?classifier=sources&repository_url=repo.spring.io/release'
        type = 'maven'  # NOQA
        namespace = 'org.apache.xmlgraphics'
        name = 'batik-anim'
        version = '1.9.1'
        qualifiers_as_dict = {
            'classifier': 'sources',
            'repository_url': 'repo.spring.io/release'
        }
        subpath = None

        purl = PackageURL(type, namespace, name, version,
                          qualifiers_as_dict, subpath)
        assert canonical_purl == purl.to_string()

    def test_normalize_encode_can_take_unicode_with_non_ascii_with_slash(self):
        uncd = u'núcleo/núcleo'
        normal = normalize(
            type=uncd, namespace=uncd, name=uncd, version=uncd,
            qualifiers='a=' + uncd, subpath=uncd, encode=True)
        expected = (
            'n%c3%bacleo/n%c3%bacleo',
            'n%C3%BAcleo/n%C3%BAcleo',
            'n%C3%BAcleo/n%C3%BAcleo',
            'n%C3%BAcleo/n%C3%BAcleo',
            'a=n%C3%BAcleo/n%C3%BAcleo',
            'n%C3%BAcleo/n%C3%BAcleo'
        )
        assert expected == normal

    def test_normalize_decode_can_take_unicode_with_non_ascii_with_slash(self):
        uncd = u'núcleo/núcleo'
        normal = normalize(
            type=uncd, namespace=uncd, name=uncd, version=uncd,
            qualifiers='a=' + uncd, subpath=uncd, encode=False)
        expected = (
            'núcleo/núcleo',
            'núcleo/núcleo',
            'núcleo/núcleo',
            'núcleo/núcleo',
            {'a': 'núcleo/núcleo'},
            'núcleo/núcleo',
        )
        assert expected == normal

    def test_normalize_encode_always_reencodes(self):
        uncd = u'n%c3%bacleo/n%c3%bacleo'
        normal = normalize(
            type=uncd, namespace=uncd, name=uncd, version=uncd,
            qualifiers='a=' + uncd, subpath=uncd, encode=True)
        expected = (
            u'n%25c3%25bacleo/n%25c3%25bacleo',
            u'n%25c3%25bacleo/n%25c3%25bacleo',
            u'n%25c3%25bacleo/n%25c3%25bacleo',
            u'n%25c3%25bacleo/n%25c3%25bacleo',
            u'a=n%25c3%25bacleo/n%25c3%25bacleo',
            u'n%25c3%25bacleo/n%25c3%25bacleo'
        )
        assert expected == normal

    def test_qualifiers_must_be_key_value_pairs(self):
        purl = 'pkg:maven/org.apache.xmlgraphics/batik-anim@1.9.1?this+is+not+a+key_value'
        try:
            PackageURL.from_string(purl)
            self.fail('Failed to raise exception for invalid qualifiers')
        except ValueError as ve:
            assert 'Invalid qualifier. Must be a string of key=value pairs' in str(ve)

    def test_to_dict_optionally_returns_qualifiers_as_string(self):
        purl = PackageURL(
            type='maven',
            namespace='org.apache',
            name='commons-logging',
            version='12.3',
            qualifiers='this=12&that=13',
            subpath='this/is/a/path',
        )

        expected = dict([
            ('type', 'maven'),
            ('namespace', 'org.apache'),
            ('name', 'commons-logging'),
            ('version', '12.3'),
            ('qualifiers', dict([
                ('that', '13'),
                ('this', '12'),
            ])),
            ('subpath', 'this/is/a/path')
        ])
        assert expected == purl.to_dict()

        expected = dict([
            ('type', u'maven'),
            ('namespace', u'org.apache'),
            ('name', u'commons-logging'),
            ('version', u'12.3'),
            ('qualifiers', u'that=13&this=12'),
            ('subpath', u'this/is/a/path')
        ])
        assert expected == purl.to_dict(encode=True)

    def test_to_dict_custom_empty_value(self):
        purl = PackageURL(
            type='maven',
            namespace='',
            name='commons-logging',
            version='12.3',
            qualifiers=None,
        )

        expected = dict([
            ('type', 'maven'),
            ('namespace', None),
            ('name', 'commons-logging'),
            ('version', '12.3'),
            ('qualifiers', None),
            ('subpath', None)
        ])
        assert expected == purl.to_dict()
        assert expected == purl.to_dict(empty=None)

        expected = dict([
            ('type', 'maven'),
            ('namespace', ''),
            ('name', 'commons-logging'),
            ('version', '12.3'),
            ('qualifiers', ''),
            ('subpath', '')
        ])
        assert expected == purl.to_dict(empty='')


def test_purl_is_hashable():
    s = {PackageURL(name='hashable', type='pypi')}
    assert len(s) == 1




# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import inspect
import re
from functools import wraps

"""
Given a URI regex (or some string), this module can route execution to a
callable.

There are several routing implementations available in Rails, Django, Flask,
Paste, etc. However, these all assume that the routed processing is to craft a
response to an incoming external HTTP request.

Here we are instead doing the opposite: given a URI (and no request yet) we are
routing the processing to emit a request externally (HTTP or other protocol)
and handling its response.

Also we crawl a lot and not only HTTP: git, svn, ftp, rsync and more.
This simple library support this kind of arbitrary URI routing.

This is inspired by Guido's http://www.artima.com/weblogs/viewpost.jsp?thread=101605
and Django, Flask, Werkzeug and other url dispatch and routing design from web
frameworks.
https://github.com/douban/brownant has a similar approach, using
Werkzeug with the limitation that it does not route based on URI scheme and is
limited to HTTP.
"""


class Rule(object):
    """
    A rule is a mapping between a pattern (typically a URI) and a callable
    (typically a function).
    The pattern is a regex string pattern and must match entirely a string
    (typically a URI) for the rule to be considered, i.e. for the endpoint to
    be resolved and eventually invoked for a given string (typically a URI).
    """
    def __init__(self, pattern, endpoint):
        # To ensure the pattern will match entirely, we wrap the pattern
        # with start of line ^ and  end of line $.
        self.pattern = pattern.lstrip('^').rstrip('$')
        self.pattern_match = re.compile('^' + self.pattern + '$').match

        # ensure the endpoint is callable
        assert callable(endpoint)
        # classes are not always callable, make an extra check
        if inspect.isclass(endpoint):
            obj = endpoint()
            assert callable(obj)

        self.endpoint = endpoint

    def __repr__(self):
        return 'Rule(r"""{}""", {}.{})'.format(
            self.pattern, self.endpoint.__module__, self.endpoint.__name__)

    def match(self, string):
        """
        Match a string with the rule pattern, return True is matching.
        """
        return self.pattern_match(string)


class RouteAlreadyDefined(TypeError):
    """
    Raised when this route Rule already exists in the route map.
    """


class NoRouteAvailable(TypeError):
    """
    Raised when there are no route available.
    """


class MultipleRoutesDefined(TypeError):
    """
    Raised when there are more than one route possible.
    """


class Router(object):
    """
    A router is:
    - a container for a route map, consisting of several rules, stored in an
     ordered dictionary keyed by pattern text
    - a way to process a route, i.e. given a string (typically a URI), find the
     correct rule and invoke its callable endpoint
    - and a convenience decorator for routed callables (either a function or
     something with a __call__ method)

    Multiple routers can co-exist as needed, such as a router to collect,
    another to fetch, etc.
    """
    def __init__(self, route_map=None):
        """
        'route_map' is an ordered mapping of pattern -> Rule.
        """
        self.route_map = route_map or dict()
        # lazy cached pre-compiled regex match() for all route patterns
        self._is_routable = None

    def __repr__(self):
        return repr(self.route_map)

    def __iter__(self):
        return iter(self.route_map.items())

    def keys(self):
        return self.route_map.keys()

    def append(self, pattern, endpoint):
        """
        Append a new pattern and endpoint Rule at the end of the map.
        Use this as an alternative to the route decorator.
        """
        if pattern in self.route_map:
            raise RouteAlreadyDefined(pattern)
        self.route_map[pattern] = Rule(pattern, endpoint)

    def route(self, *patterns):
        """
        Decorator to make a callable 'endpoint' routed to one or more patterns.

        Example:
        >>> my_router = Router()
        >>> @my_router.route('http://nexb.com', 'http://deja.com')
        ... def somefunc(uri):
        ...    pass
        """
        def decorator(endpoint):
            assert patterns
            for pat in patterns:
                self.append(pat, endpoint)

            @wraps(endpoint)
            def decorated(*args, **kwargs):
                return self.process(*args, **kwargs)
            return decorated

        return decorator

    def process(self, string, *args, **kwargs):
        """
        Given a string (typically a URI), resolve this string to an endpoint
        by searching available rules then execute the endpoint callable for
        that string passing down all arguments to the endpoint invocation.
        """
        endpoint = self.resolve(string)
        if inspect.isclass(endpoint):
            # instantiate a class, that must define a __call__ method
            # TODO: consider passing args to the constructor?
            endpoint = endpoint()
        # call the callable
        return endpoint(string, *args, **kwargs)

    def resolve(self, string):
        """
        Resolve a string: given a string (typically a URI) resolve and
        return the best endpoint function for that string.

        Ambiguous resolution is not allowed in order to keep things in
        check when there are hundreds rules: if multiple routes are
        possible for a string (typically a URI), a MultipleRoutesDefined
        TypeError is raised.
        """
        # TODO: we could improve the performance of this by using a single
        # regex and named groups if this ever becomes a bottleneck.
        candidates = [r for r in self.route_map.values() if r.match(string)]

        if not candidates:
            raise NoRouteAvailable(string)

        if len(candidates) > 1:
            # this can happen when multiple patterns match the same string
            # we raise an exception with enough debugging information
            pats = repr([r.pattern for r in candidates])
            msg = '%(string)r matches multiple patterns %(pats)r' % locals()
            raise MultipleRoutesDefined(msg)

        return candidates[0].endpoint

    def is_routable(self, string):
        """
        Return True if `string` is routable by this router, e.g. if it
        matches any of the route patterns.
        """
        if not string:
            return

        if not self._is_routable:
            # build an alternation regex
            routables = '^(' + '|'.join(pat for pat in self.route_map) + ')$'
            self._is_routable = re.compile(routables, re.UNICODE).match

        return bool(self._is_routable(string))



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import string
from collections import namedtuple
from urllib.parse import quote as _percent_quote
from urllib.parse import unquote as _percent_unquote
from urllib.parse import urlsplit as _urlsplit

# Python 3
unicode = str  # NOQA
basestring = (bytes, str,)  # NOQA

"""
A purl (aka. Package URL) implementation as specified at:
https://github.com/package-url/purl-spec
"""


def quote(s):
    """
    Return a percent-encoded unicode string, except for colon :, given an `s`
    byte or unicode string.
    """
    if isinstance(s, unicode):
        s = s.encode('utf-8')
    quoted = _percent_quote(s)
    if not isinstance(quoted, unicode):
        quoted = quoted.decode('utf-8')
    quoted = quoted.replace('%3A', ':')
    return quoted


def unquote(s):
    """
    Return a percent-decoded unicode string, given an `s` byte or unicode
    string.
    """
    unquoted = _percent_unquote(s)
    if not isinstance(unquoted, unicode):
        unquoted = unquoted .decode('utf-8')
    return unquoted


def get_quoter(encode=True):
    """
    Return quoting callable given an `encode` tri-boolean (True, False or None)
    """
    if encode is True:
        return quote
    elif encode is False:
        return unquote
    elif encode is None:
        return lambda x: x


def normalize_type(type, encode=True):  # NOQA
    if not type:
        return
    if not isinstance(type, unicode):
        type = type.decode('utf-8')  # NOQA

    quoter = get_quoter(encode)
    type = quoter(type)  # NOQA
    return type.strip().lower() or None


def normalize_namespace(namespace, ptype, encode=True):  # NOQA
    if not namespace:
        return
    if not isinstance(namespace, unicode):
        namespace = namespace.decode('utf-8')

    namespace = namespace.strip().strip('/')
    if ptype in ('bitbucket', 'github', 'pypi', 'gitlab'):
        namespace = namespace.lower()
    segments = [seg for seg in namespace.split('/') if seg.strip()]
    segments = map(get_quoter(encode), segments)
    return '/'.join(segments) or None


def normalize_name(name, ptype, encode=True):  # NOQA
    if not name:
        return
    if not isinstance(name, unicode):
        name = name.decode('utf-8')

    quoter = get_quoter(encode)
    name = quoter(name)
    name = name.strip().strip('/')
    if ptype in ('bitbucket', 'github', 'pypi', 'gitlab'):
        name = name.lower()
    if ptype in ('pypi',):
        name = name.replace('_', '-')
    return name or None


def normalize_version(version, encode=True):  # NOQA
    if not version:
        return
    if not isinstance(version, unicode):
        version = version.decode('utf-8')

    quoter = get_quoter(encode)
    version = quoter(version.strip())
    return version or None


def normalize_qualifiers(qualifiers, encode=True):  # NOQA
    """
    Return normalized `qualifiers` as a mapping (or as a string if `encode` is
    True). The `qualifiers` arg is either a mapping or a string.
    Always return a mapping if decode is True (and never None).
    Raise ValueError on errors.
    """
    if not qualifiers:
        return None if encode else dict()

    if isinstance(qualifiers, basestring):
        if not isinstance(qualifiers, unicode):
            qualifiers = qualifiers.decode('utf-8')
        # decode string to list of tuples
        qualifiers = qualifiers.split('&')
        if not all('=' in kv for kv in qualifiers):
            raise ValueError(
                'Invalid qualifier. '
                'Must be a string of key=value pairs:{}'.format(repr(qualifiers)))
        qualifiers = [kv.partition('=') for kv in qualifiers]
        qualifiers = [(k, v) for k, _, v in qualifiers]
    elif isinstance(qualifiers, dict):
        qualifiers = qualifiers.items()
    else:
        raise ValueError(
            'Invalid qualifier. '
            'Must be a string or dict:{}'.format(repr(qualifiers)))

    quoter = get_quoter(encode)
    qualifiers = {k.strip().lower(): quoter(v)
        for k, v in qualifiers if k and k.strip() and v and v.strip()}

    valid_chars = string.ascii_letters + string.digits + '.-_'
    for key in qualifiers:
        if not key:
            raise ValueError('A qualifier key cannot be empty')

        if '%' in key:
            raise ValueError(
                "A qualifier key cannot be percent encoded: {}".format(repr(key)))

        if ' ' in key:
            raise ValueError(
                "A qualifier key cannot contain spaces: {}".format(repr(key)))

        if not all(c in valid_chars for c in key):
            raise ValueError(
                "A qualifier key must be composed only of ASCII letters and numbers"
                "period, dash and underscore: {}".format(repr(key)))

        if key[0] in string.digits:
            raise ValueError(
                "A qualifier key cannot start with a number: {}".format(repr(key)))

    qualifiers = sorted(qualifiers.items())
    qualifiers = dict(qualifiers)
    if encode:
        qualifiers = ['{}={}'.format(k, v) for k, v in qualifiers.items()]
        qualifiers = '&'.join(qualifiers)
        return qualifiers or None
    else:
        return qualifiers or {}


def normalize_subpath(subpath, encode=True):  # NOQA
    if not subpath:
        return None
    if not isinstance(subpath, unicode):
        subpath = subpath.decode('utf-8')

    quoter = get_quoter(encode)
    segments = subpath.split('/')
    segments = [quoter(s) for s in segments if s.strip() and s not in ('.', '..')]
    subpath = '/'.join(segments)
    return subpath or None


def normalize(type, namespace, name, version, qualifiers, subpath, encode=True):  # NOQA
    """
    Return normalized purl components
    """
    type = normalize_type(type, encode)  # NOQA
    namespace = normalize_namespace(namespace, type, encode)
    name = normalize_name(name, type, encode)
    version = normalize_version(version, encode)
    qualifiers = normalize_qualifiers(qualifiers, encode)
    subpath = normalize_subpath(subpath, encode)
    return type, namespace, name, version, qualifiers, subpath


_components = ['type', 'namespace', 'name', 'version', 'qualifiers', 'subpath']


class PackageURL(namedtuple('PackageURL', _components)):
    """
    A purl is a package URL as defined at
    https://github.com/package-url/purl-spec
    """

    def __new__(self, type=None, namespace=None, name=None,  # NOQA
                version=None, qualifiers=None, subpath=None):

        required = dict(type=type, name=name)
        for key, value in required.items():
            if value:
                continue
            raise ValueError('Invalid purl: {} is a required argument.'
                             .format(key))

        strings = dict(
            type=type,
            namespace=namespace,
            name=name,
            version=version,
            subpath=subpath,
        )

        for key, value in strings.items():
            if value and isinstance(value, basestring) or not value:
                continue
            raise ValueError('Invalid purl: {} argument must be a string: {}.'
                             .format(key, repr(value)))

        if qualifiers and not isinstance(qualifiers, (basestring, dict,)):
            raise ValueError('Invalid purl: {} argument must be a dict or a string: {}.'
                             .format('qualifiers', repr(qualifiers)))

        type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
            type, namespace, name, version, qualifiers, subpath, encode=None)

        return super(PackageURL, self).__new__(
            PackageURL,
            type=type,
            namespace=namespace,
            name=name,
            version=version,
            qualifiers=qualifiers,
            subpath=subpath,
        )

    def __str__(self, *args, **kwargs):
        return self.to_string()

    def __hash__(self):
        return hash(self.to_string())

    def to_dict(self, encode=False, empty=None):
        """
        Return an ordered dict of purl components as {key: value}.
        If `encode` is True, then "qualifiers" are encoded as a normalized
        string. Otherwise, qualifiers is a mapping.
        You can provide a value for `empty` to be used in place of default None.
        """
        data = self._asdict()
        if encode:
            data['qualifiers'] = normalize_qualifiers(self.qualifiers, encode=encode)

        for field, value in data.items():
            data[field] = value or empty

        return data

    def to_string(self):
        """
        Return a purl string built from components.
        """
        type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
            self.type,
            self.namespace,
            self.name,
            self.version,
            self.qualifiers,
            self.subpath,
            encode=True,
        )

        purl = ['pkg:', type, '/']

        if namespace:
            purl.append(namespace)
            purl.append('/')

        purl.append(name)

        if version:
            purl.append('@')
            purl.append(version)

        if qualifiers:
            purl.append('?')
            purl.append(qualifiers)

        if subpath:
            purl.append('#')
            purl.append(subpath)

        return ''.join(purl)

    @classmethod
    def from_string(cls, purl):
        """
        Return a PackageURL object parsed from a string.
        Raise ValueError on errors.
        """
        if (not purl or not isinstance(purl, basestring)
            or not purl.strip()):
            raise ValueError('A purl string argument is required.')

        scheme, sep, remainder = purl.partition(':')
        if not sep or scheme != 'pkg':
            raise ValueError(
                'purl is missing the required '
                '"pkg" scheme component: {}.'.format(repr(purl)))

        # this strip '/, // and /// as possible in :// or :///
        remainder = remainder.strip().lstrip('/')

        type, sep, remainder = remainder.partition('/')  # NOQA
        if not type or not sep:
            raise ValueError(
                'purl is missing the required '
                'type component: {}.'.format(repr(purl)))

        scheme, authority, path, qualifiers, subpath = _urlsplit(
            url=remainder, scheme='', allow_fragments=True)

        if scheme or authority:
            msg = ('Invalid purl {} cannot contain a "user:pass@host:port" '
                   'URL Authority component: {}.')
            raise ValueError(msg.format(repr(purl), repr(authority)))

        path = path.lstrip('/')
        remainder, sep, version = path.rpartition('@')
        if not sep:
            remainder = version
            version = None

        ns_name = remainder.strip().strip('/')
        ns_name = ns_name.split('/')
        ns_name = [seg for seg in ns_name if seg and seg.strip()]
        namespace = ''
        name = ''
        if len(ns_name) > 1:
            name = ns_name[-1]
            ns = ns_name[0:-1]
            namespace = '/'.join(ns)
        elif len(ns_name) == 1:
            name = ns_name[0]

        if not name:
            raise ValueError(
                'purl is missing the required '
                'name component: {}'.format(repr(purl)))

        type, namespace, name, version, qualifiers, subpath = normalize(# NOQA
            type,
            namespace,
            name,
            version,
            qualifiers,
            subpath,
            encode=False,
        )

        return PackageURL(type, namespace, name, version, qualifiers, subpath)



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import django_filters


class PackageURLFilter(django_filters.CharFilter):
    """
    Filter by a Package URL string.
    Empty values are not applied to the QuerySet.

    This filter depends on a `for_package_url` method available on the Model
    Manager, see for example `PackageURLQuerySetMixin`.
    """
    def filter(self, qs, value):
        empty_values = ([], (), {}, '', None)
        if value in empty_values:
            return qs

        if self.distinct:
            qs = qs.distinct()
        return qs.for_package_url(value)



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import io
import json
import os
import re
from unittest import TestCase

from packageurl.contrib.url2purl import get_purl as purl_getter


def get_purl(url):
    purl = purl_getter(url)
    return purl and str(purl.to_string())


class TestURL2PURL(TestCase):

    def test_get_purl_empty_string(self):
        self.assertEqual(None, get_purl(''))

    def test_get_purl_none(self):
        self.assertEqual(None, get_purl(None))

    def test_get_purl_unroutable_uri(self):
        self.assertEqual(None, get_purl('dsf.example'))


def python_safe(s):
    """
    Return a name safe to use as a python function name.
    """
    safe_chars = re.compile(r'[\W_]', re.MULTILINE)
    s = s.strip().lower()
    s = [x for x in safe_chars.split(s) if x]
    return '_'.join(s)


def get_url2purl_test_method(test_url, expected_purl):

    def test_method(self):
        self.assertEqual(expected_purl, get_purl(test_url))

    return test_method


def build_tests(clazz, test_file='url2purl.json', regen=False):
    """
    Dynamically build test methods for Package URL inference from a JSON test
    file.
    The JSON test file is a key-sorted mapping of {test url: expected purl}.
    """
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')
    test_file = os.path.join(test_data_dir, test_file)

    with io.open(test_file, encoding='utf-8') as tests:
        tests_data = json.load(tests)

    if regen:
        tests_data = {test_url: get_purl(test_url)
                      for test_url in tests_data.keys()}
        with io.open(test_file, 'w') as regened:
            json.dump(tests_data, regened, indent=2)

    for test_url, expected_purl in sorted(tests_data.items()):
        test_name = 'test_url2purl_{test_url}'.format(test_url=test_url)
        test_name = python_safe(test_name)
        test_method = get_url2purl_test_method(test_url, expected_purl)
        test_method.funcname = test_name
        # attach that method to our test class
        setattr(clazz, test_name, test_method)


class TestURL2PURLDataDriven(TestCase):
    pass


build_tests(clazz=TestURL2PURLDataDriven, regen=False)



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

import pytest

from packageurl.contrib.purl2url import purl2url


def test_purl2url_with_valid_purls():
    purls_url = {
        "pkg:github/tg1999/fetchcode": "https://github.com/tg1999/fetchcode",
        "pkg:github/tg1999/fetchcode@master": "https://github.com/tg1999/fetchcode/tree/master",
        "pkg:github/tg1999/fetchcode@master#tests": "https://github.com/tg1999/fetchcode/tree/master/tests",
        "pkg:github/tg1999": None,
        "pkg:cargo/clap@2.3.3": "https://crates.io/api/v1/crates/clap/2.3.3/download",
        "pkg:cargo/rand@0.7.2": "https://crates.io/api/v1/crates/rand/0.7.2/download",
        "pkg:cargo/structopt@0.3.11": "https://crates.io/api/v1/crates/structopt/0.3.11/download",
        "pkg:cargo/abc": None,
        "pkg:rubygems/unf@0.1.3": "https://rubygems.org/downloads/unf-0.1.3.gem",
        "pkg:rubygems/yajl-ruby@1.2.0": "https://rubygems.org/downloads/yajl-ruby-1.2.0.gem",
        "pkg:gem/package-name": None,
        "pkg:bitbucket/birkenfeld/pygments-main": "https://bitbucket.org/birkenfeld/pygments-main",
        "pkg:bitbucket/birkenfeld/pygments-main@244fd47e07d1014f0aed9c": "https://bitbucket.org/birkenfeld/pygments-main/src/244fd47e07d1014f0aed9c",
        "pkg:bitbucket/birkenfeld/pygments-main@master#views": "https://bitbucket.org/birkenfeld/pygments-main/src/master/views",
        "pkg:bitbucket/birkenfeld": None,
        "pkg:gitlab/tg1999/firebase@master": "https://gitlab.com/tg1999/firebase/-/tree/master",
        "pkg:gitlab/tg1999/firebase@1a122122#views": "https://gitlab.com/tg1999/firebase/-/tree/1a122122/views",
        "pkg:gitlab/tg1999/firebase": "https://gitlab.com/tg1999/firebase",
        "pkg:gitlab/tg1999": None,
    }

    for purl, url in purls_url.items():
        assert url == purl2url(purl)


def test_convert_with_invalid_purls():
    purls = ["pkg:github", "pkg:cargo", "pkg:gem", "pkg:bitbucket", "pkg:gitlab", None]
    with pytest.raises(Exception) as e_info:
        for purl in purls:
            url = purl2url(purl)
            assert "Invalid PURL" == e_info



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

from packageurl import PackageURL
from packageurl.contrib.route import NoRouteAvailable
from packageurl.contrib.route import Router

router = Router()


def purl2url(purl):
    """
    Return a URL inferred from the `purl` string
    """
    if purl:
        try:
            return router.process(purl)
        except NoRouteAvailable:
            return


get_url = purl2url


@router.route("pkg:cargo/.*")
def build_cargo_download_url(purl):
    """
    Return a cargo download URL `url` from a the `purl` string
    """
    purl_data = PackageURL.from_string(purl)

    name = purl_data.name
    version = purl_data.version

    if not (name and version):
        return

    return "https://crates.io/api/v1/crates/{name}/{version}/download".format(
        name=name, version=version
    )


@router.route("pkg:bitbucket/.*")
def build_bitbucket_homepage_url(purl):
    """
    Return a bitbucket homepage URL `url` from a the `purl` string
    """
    purl_data = PackageURL.from_string(purl)

    namespace = purl_data.namespace
    name = purl_data.name
    version = purl_data.version
    subpath = purl_data.subpath

    if not (name and namespace):
        return

    url = "https://bitbucket.org/{namespace}/{name}".format(
        namespace=namespace, name=name
    )
    if version:
        url = "{url}/src/{version}".format(url=url, version=version)

    if subpath:
        url = "{url}/{subpath}".format(url=url, subpath=subpath)

    return url


@router.route("pkg:github/.*")
def build_github_homepage_url(purl):
    """
    Return a github homepage URL `url` from a the `purl` string
    """
    purl_data = PackageURL.from_string(purl)

    namespace = purl_data.namespace
    name = purl_data.name
    version = purl_data.version
    subpath = purl_data.subpath

    if not (name and namespace):
        return

    url = "https://github.com/{namespace}/{name}".format(namespace=namespace, name=name)

    if version:
        url = "{url}/tree/{version}".format(url=url, version=version)

    if subpath:
        url = "{url}/{subpath}".format(url=url, subpath=subpath)

    return url


@router.route("pkg:gitlab/.*")
def build_gitlab_homepage_url(purl):
    """
    Return a gitlab homepage URL `url` from a the `purl` string
    """
    purl_data = PackageURL.from_string(purl)

    namespace = purl_data.namespace
    name = purl_data.name
    version = purl_data.version
    subpath = purl_data.subpath

    if not (name and namespace):
        return

    url = "https://gitlab.com/{namespace}/{name}".format(namespace=namespace, name=name)

    if version:
        url = "{url}/-/tree/{version}".format(url=url, version=version)

    if subpath:
        url = "{url}/{subpath}".format(url=url, subpath=subpath)

    return url


@router.route("pkg:rubygems/.*")
def build_gem_download_url(purl):
    """
    Return a rubygems homepage URL `url` from a the `purl` string
    """
    purl_data = PackageURL.from_string(purl)

    name = purl_data.name
    version = purl_data.version

    if not (name and version):
        return

    return "https://rubygems.org/downloads/{name}-{version}.gem".format(
        name=name, version=version
    )



# -*- coding: utf-8 -*-
#
# Copyright (c) the purl authors
# SPDX-License-Identifier: MIT
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Visit https://github.com/package-url/packageurl-python for support and
# download.

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from packageurl import PackageURL


def without_empty_values(input_dict):
    """
    Return a new dict not including empty value entries from `input_dict`.

    `None`, empty string, empty list, and empty dict/set are cleaned.
    `0` and `False` values are kept.
    """
    empty_values = ([], (), {}, '', None)

    return {
        key: value
        for key, value in input_dict.items()
        if value not in empty_values
    }


def purl_to_lookups(purl):
    """
    Return a lookups dict built from the provided `purl` string.
    Those lookups can be used as QuerySet filters.
    """
    if not purl.startswith('pkg:'):
        purl = 'pkg:' + purl

    try:
        package_url = PackageURL.from_string(purl)
    except ValueError:
        return  # Not a valid PackageURL

    package_url_dict = package_url.to_dict(encode=True)
    return without_empty_values(package_url_dict)


class PackageURLQuerySetMixin:
    """
    Add Package URL filtering method to a django.db.models.QuerySet.
    """
    def for_package_url(self, purl_str):
        """
        Filter the QuerySet with the provided Package URL string.
        The purl string is validated and transformed into filtering lookups.
        """
        lookups = purl_to_lookups(purl_str)
        if lookups:
            return self.filter(**lookups)
        return self.none()


class PackageURLQuerySet(PackageURLQuerySetMixin, models.QuerySet):
    pass


class PackageURLMixin(models.Model):
    """
    Abstract Model for Package URL "purl" fields support.
    """
    type = models.CharField(
        max_length=16,
        blank=True,
        help_text=_(
            'A short code to identify the type of this package. '
            'For example: gem for a Rubygem, docker for a container, '
            'pypi for a Python Wheel or Egg, maven for a Maven Jar, '
            'deb for a Debian package, etc.'
        )
    )

    namespace = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            'Package name prefix, such as Maven groupid, Docker image owner, '
            'GitHub user or organization, etc.'
        ),
    )

    name = models.CharField(
        max_length=100,
        blank=True,
        help_text=_('Name of the package.'),
    )

    version = models.CharField(
        max_length=100,
        blank=True,
        help_text=_('Version of the package.'),
    )

    qualifiers = models.CharField(
        max_length=1024,
        blank=True,
        help_text=_(
            'Extra qualifying data for a package such as the name of an OS, '
            'architecture, distro, etc.'
        ),
    )

    subpath = models.CharField(
        max_length=200,
        blank=True,
        help_text=_(
            'Extra subpath within a package, relative to the package root.'
        ),
    )

    objects = PackageURLQuerySet.as_manager()

    class Meta:
        abstract = True

    @property
    def package_url(self):
        """
        Return a compact Package URL "purl" string.
        """
        try:
            purl = PackageURL(
                self.type, self.namespace, self.name,
                self.version, self.qualifiers, self.subpath
            )
        except ValueError:
            return ''
        return str(purl)

    def set_package_url(self, package_url):
        """
        Set each field values to the values of the provided `package_url` string
        or PackageURL object.
        Existing values are always overwritten, forcing the new value or an
        empty string on all the `package_url` fields since we do not want to
        keep any previous values.
        """
        if not isinstance(package_url, PackageURL):
            package_url = PackageURL.from_string(package_url)

        package_url_dict = package_url.to_dict(encode=True, empty='')
        for field_name, value in package_url_dict.items():
            model_field = self._meta.get_field(field_name)

            if value and len(value) > model_field.max_length:
                message = _('Value too long for field "{}".'.format(field_name))
                raise ValidationError(message)

            setattr(self, field_name, value)
